# 	Copyright (c) 2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from command import command
from cxutil import flags,pmask,cmask

from cxutil.UID import UID
from cxutil.utils import opts
from cxutil.exceptions import BranchPass
from threading import Thread,Lock

from cxcore.cxnetlink import cl_mutex
import traceback

CL_DOMAIN = "_cxheartbeat._udp.local"

def acquire(signal,mutex,cell,name):
	res = mutex.acquire()
	cell["status"] = res
	cell["thread"] = None
	signal(("mutex","mutex.%s.lock" % (name)), {"res": res})


class Default (command):
	'''
	Use tree branch in active/passive cluster configuration
	'''

	name = "ap-cluster"
	mask = flags.Begin | flags.Transparent | flags.Esoteric | flags.Internal | flags.LocalsTransparent

	parameters = {
		"$1": {
			"type": "word",
			"description": "Cluster name",
			"name": "name",
		},
	}

	def pre(self, argv):

		command.pre(self,argv)
		if argv.call_mask & cmask.Immediate:
			self.uid = UID()
			sig = "mutex.%s.lock" % (argv.envl.name)
			self.catch = (sig,)
			self.signal_keys[sig] = cmask.ScriptUp
			self.lock = Lock()
			return

		# if there is an acquired mutex, run `postUp()` through all subtree
		# if there is a registered (not acquired) mutex, do nothing
		# if there is no registered mutex, run registration thread
		
		if not self.lock.acquire(False):
			return

		try:
			mx = argv.envg.mx[argv.envl.name]
		except:
			if not hasattr(argv.envg,"mx"):
				argv.envg.mx = {}
			
			mutex = cl_mutex(argv.esoteric.bus, "%s.%s." % (argv.envl.name, CL_DOMAIN), "%s." % (CL_DOMAIN))
			mx = argv.envg.mx[argv.envl.name] = opts({
				"instance": mutex,
				"thread": None,
				"status": None,
				"used": [self.uid,]
			})
			
		if not self.uid in mx.used:
			mx.used.append(self.uid)


		###
		#
		# Start mutex acquire thread, if there is _no_ thread, _no_ status and there _is_ mutex instance.
		# Then just skip branch.
		#
		###
		if \
			(not mx.thread) and \
			(not mx.status) and \
			mx.instance:

			t = Thread(target=acquire,args=(self.signal,mx.instance,argv.envg.mx[argv.envl.name],argv.envl.name))
			mx.thread = t
			t.start()
			self.lock.release()
			raise BranchPass
		
		###
		#
		# If there is the thread, the instance, but _no_ status, skip branch and wait for a signal
		#
		###
		if \
			mx.thread and \
			(not mx.status) and \
			mx.instance:
			
			self.lock.release()
			raise BranchPass
		
		###
		#
		# Ok, let's go
		#
		###
		self.lock.release()

	def terminate(self,argv):
		
		mx = argv.envg.mx[argv.envl.name]
		if self.uid in mx.used:
			mx.used.remove(self.uid)
		
		if not mx.used:
			mx.instance.release()
			del argv.envg.mx[argv.envl.name]
			del mx
